(* from  "Acyclic Strategy  for Silent Self-Stabilization  in Spanning
   Forests" Karine Altisen, Stéphane Devismes,  and Anaı̈s Durand,
   SSS 2018

https://hal.archives-ouvertes.fr/hal-01938671 *)

open Algo
open State

let init_state (i: int) (id: string): State.t =
  {
    pid = id;
    input = 0;
    sub = 0;
    res = 0
  }

let children_states pid (neighbors: 'st neighbor list): 'st list =
  let n =
    match Algo.parent pid with
    | None -> neighbors
    | Some par -> List.filteri (fun i _ -> i <> par) neighbors
  in
  List.map state n

let parent_opt pid (nl: State.t neighbor list): State.t option =
  let parent_i = Algo.parent pid in
  Option.map (fun par -> state (List.nth nl par)) parent_i

let parent_state (s: State.t) (nl: State.t neighbor list): State.t =
  match parent_opt s.pid nl with
  | Some par -> par
  | None ->
    let nls = String.concat " " (List.map (fun s -> (state s).pid) nl) in
    Printf.eprintf "No parent found for pid=%s, nl=%s\n" s.pid nls;
    assert false

let enable_f (p: State.t) (nl: State.t neighbor list): action list =
  let children = children_states p.pid nl in
  let sum_subs = List.fold_left (+) 0 (List.map (fun s -> s.sub) children) in
  if p.sub <> p.input + sum_subs then ["S"]
  else
    if p.pid = "root" then
      if (p.res <> p.sub) then ["Rr"] else []
    else
      if p.res <> (max p.sub (parent_state p nl).res) then ["Rp"] else []

let step_f (p: State.t) (nl: State.t neighbor list) (a: action): State.t =
  match a with
  | "S" ->
    let children = children_states p.pid nl in
    let sum_subs = List.fold_left (+) 0 (List.map (fun s -> s.sub) children) in
    {p with sub = p.input + sum_subs}

  | "Rr" ->
    {p with res = p.sub}

  | "Rp" ->
    {p with res = max p.sub (parent_state p nl).res}

  | _ ->
    assert false
